/****************************************************************************
    
    This file is part of The Newtonator soft synth.

    Copyright (C) 2010-2011 Michael Bechard <mbechard@users.sourceforge.net>

    The Newtonator is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The Newtonator is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Newtonator.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************/

#include "NewtonatorInstr.h"
#include <stk/Skini.h>

using namespace stk;
using namespace std;

const int NewtonatorInstr::NUM_NOTES = 6;
const size_t NewtonatorInstr::POLYPHONY = 5;

NewtonatorInstr::NewtonatorInstr(unsigned int channels)
{
    for (int note = 0; note < NUM_NOTES; note++)
    {
        //_newtonators.push_back(new Newtonator(channels));
        Newtonator* newNewt = new Newtonator(true, channels);
        for (unsigned int chan = 0; chan < channels; chan++)
            newNewt->getFloorSources()[chan].setFrequency(Midi2Pitch[note]);
        //_newtonators.insert(NewtCollection::value_type(Midi2Pitch[note], newNewt));
        _newtonators.push_back(newNewt);
    }
    _currNewtIt = _newtonators.begin();
    addSampleRateAlert(this);
    initChannels(channels);
}

NewtonatorInstr::~NewtonatorInstr()
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        delete *(it++);
    }
}

void NewtonatorInstr::setAllAmpEnvTimes(StkFloat aTime, StkFloat dTime, StkFloat sLevel, StkFloat rTime)
{
    NewtCollection::iterator it = _newtonators.begin();
    ADSR* ampEnvs = NULL;
    unsigned int numChannels = channelsOut();

    while (it != _newtonators.end())
    {
        ampEnvs = (*(it++))->getAmpEnvs();
        for (unsigned int chan = 0; chan < numChannels; chan++)
        {
            ampEnvs[chan].setAllTimes(aTime, dTime, sLevel, rTime);
        }
    }
}

void NewtonatorInstr::setAllGModEnvTimes(StkFloat aTime, StkFloat dTime, StkFloat sLevel, StkFloat rTime)
{
    NewtCollection::iterator it = _newtonators.begin();
    VariSource* gmodSources = NULL;
    unsigned int numChannels = channelsOut();

    while (it != _newtonators.end())
    {
        gmodSources = (*(it++))->getGravModSources();
        for (unsigned int chan = 0; chan < numChannels; chan++)
        {
            gmodSources[chan].getEnv().setAllTimes(aTime, dTime, sLevel, rTime);
        }
    }
}

StkFloat NewtonatorInstr::tick(unsigned int channel)
{
    StkFloat ret = 0.0;
    PolyphonyQueue::iterator it = _activeNewts.begin();

    while (it != _activeNewts.end())
    {
        if ((*it)->isPlaying())
        {
            ret += (*it)->newtTick(channel, 0);
            it++;
        }
        else
        {
            it = _activeNewts.erase(it);
        }
    }

    //ret = _currNewt->newtTick(channel);

    return ret;
}

void NewtonatorInstr::sampleRateChanged(StkFloat newRate, StkFloat oldRate)
{
    //May need to do something here, I dunno yet
}

void NewtonatorInstr::initChannels(unsigned int numChannels)
{
    //Make sure the number is reflected in lastFrame_
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->initChannels(numChannels);
    }

    lastFrame_.resize(1, numChannels, 0.0);
}

void NewtonatorInstr::addActiveNewt(Newtonator* newt)
{
    _currNewt = newt;
    _activeNewts.push_back(newt);
    if (_activeNewts.size() > POLYPHONY)
    {
        _activeNewts.pop_front();
    }
}

void NewtonatorInstr::noteOn(StkFloat frequency, StkFloat amplitude)
{
    if (amplitude == 0.0)
    {
        //Actually a note-off event
        noteOff(frequency, .5);
    }
    else
    {
        //_newtonators[frequency]->keyOn(amplitude);
        Newtonator* newt = getNextAvailNewt();
        VariSource* floors = newt->getFloorSources();

        for (unsigned int chan = 0; chan < channelsOut(); chan++)
        {
            floors[chan].setFrequency(frequency);
        }

        newt->keyOn(amplitude);
        addActiveNewt(newt);
    }
}

void NewtonatorInstr::noteOff(StkFloat amplitude)
{
    //Uh-oh, no idea which note is being turned off. Turn off the most
    //recent "on" note
    PolyphonyQueue::reverse_iterator it(_activeNewts.end());
    PolyphonyQueue::reverse_iterator itLast(_activeNewts.begin());

    if (it != itLast)
    {
        (*it)->keyOff(amplitude);
    }
}

void NewtonatorInstr::noteOff(StkFloat frequency, StkFloat amplitude)
{
    //Turn off the most recent "on" note that matches the
    //specified frequency
    PolyphonyQueue::reverse_iterator it(_activeNewts.end());
    PolyphonyQueue::reverse_iterator itLast(_activeNewts.begin());

    while (it != itLast)
    {
        if ((*it)->getFloorSources()[0].getFrequency() == frequency)
        {
            (*it)->keyOff(amplitude);
            break;
        }
        it++;
    }
}

void NewtonatorInstr::allOff()
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->keyOff(.5);
    }
}

Newtonator* NewtonatorInstr::getNextAvailNewt()
{
    if (_currNewtIt == _newtonators.end())
        _currNewtIt = _newtonators.begin();

    return *(_currNewtIt++);
}

void NewtonatorInstr::setFrequency(StkFloat frequency)
{
    //Well, since each noteOn includes the note's frequency, there's
    //no need to implement this... yet.
}

void NewtonatorInstr::controlChange(int number, StkFloat value)
{
}

void NewtonatorInstr::setGravityScale(stk::StkFloat val)
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->setGravityScale(val);
    }
}

void NewtonatorInstr::setGravity(stk::StkFloat val)
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->setGravity(val);
    }
}

void NewtonatorInstr::setFloorAttraction(stk::StkFloat val)
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->setFloorAttraction(val);
    }
}

void NewtonatorInstr::setFloorScale(stk::StkFloat val)
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->setFloorScale(val);
    }
}

void NewtonatorInstr::setClip(bool val)
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->setClip(val);
    }
}

void NewtonatorInstr::setGravModDepth(stk::StkFloat val)
{
    NewtCollection::iterator it = _newtonators.begin();

    while (it != _newtonators.end())
    {
        (*(it++))->setGravModDepth(val);
    }
}

void NewtonatorInstr::setFloorType(VariSource::Type type)
{
    NewtCollection::iterator it = _newtonators.begin();
    unsigned int numChannels = channelsOut();
    VariSource* floorSources = NULL;

    while (it != _newtonators.end())
    {
        floorSources = (*(it++))->getFloorSources();
        for (unsigned int chan = 0; chan < numChannels; chan++)
        {
            floorSources[chan].setType(type);
        }
    }
}

void NewtonatorInstr::setGravModType(VariSource::Type type)
{
}
